/*
 * AT91SAM7 Programmer, http://www.pjrc.com/arm/sam7_pgm
 * Copyright (c) 2005, PJRC.COM, LLC, <paul@pjrc.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include "serial.h"
#include "cmd.h"
#include "settings.h"
#include "ihex.h"

#define USAGE "sam7_pgm <firmware.hex>"

void die(const char *str)
{
	printf("%s\n", str);
	exit(1);
}

#define DBGU_BRGR	0xFFFFF220	/* Baud rate setting for debug uart */
#define MC_FMR		0xFFFFFF60	/* Flash Mode Register */
#define MC_FCR		0xFFFFFF64	/* Flash Command Register */
#define MC_FSR		0xFFFFFF68	/* Flash Status Register */


int wait_flash_ready(void)
{
	int retry=0, r;
	unsigned int status;

	do {
		r = read_word(MC_FSR, &status);
		if (r) die("error while readying flash status");
		usleep(10000);
		if ((status & 1) == 1) return 0;
	} while (++retry < 100);
	return -1;
}

int unlock_flash_pages(int mask, double mclk)
{
	int page, clks, r;
	unsigned int mode;

	clks = (int)ceil(mclk * 1.5e-6);
	printf("%d clocks per 1.5 us\n", clks);
	mode = ((clks & 255) << 16) | 0x00000100;
	write_word(MC_FMR, mode);

	for (page=0; page<16; page++) {
		if (((mask >> page) & 1) == 0) continue;
		printf("Unlocking page #%d\n", page);
		r = wait_flash_ready();
		if (r) die("Timeout waiting for flash ready");
		r = write_word(MC_FCR, (page << 8) | 0x5A000004);
		if (r) die("error writing flash command");
	}
	r = wait_flash_ready();
	if (r) die("Timeout waiting for flash ready");
	return 0;
}


unsigned int word_from_firmware(unsigned int addr)
{
	return (((firmware_image[addr+3] & 255) << 24) |
		((firmware_image[addr+2] & 255) << 16) |
		((firmware_image[addr+1] & 255) << 8) |
		((firmware_image[addr+0] & 255)));
}


int main(int argc, char **argv)
{
	int r, i;
	char buf[256];
	unsigned int word, baud, lockmask, ramarea, flasharea;
	double mclk;
	int block, numblocks, blocksize, clks;

	if (argc < 2) die(USAGE);
	r = read_intel_hex(argv[1]);
	if (r < 0) die("Error reading intel hex file");
	printf("read %d bytes from %s\n", r, argv[1]);

	printf("opening serial port %s\n", port_setting());
	open_serial_port(port_setting());

	r = establish_comm_uart();
	if (r) die("Unable to sync with boot agent");

	r = bootloader_version(buf, sizeof(buf));
	if (r) die("Unable to read boot loader version");
	// todo: do something useful like showing this info to the user

	// figure out what speed the CPU is running at
	r = read_word(DBGU_BRGR, &baud);
	if (r) die("Unable to read from chip");
	mclk = 115200.0 * 16.0 * (double)baud;
	printf("MCLK is approx %.3f MHz\n", mclk / 1e6);
	if (mclk < 19e6) printf("WARNING: below 19 MHz may be unreliable\n");
	if (mclk < 1e6) die("MCLK must be at least 1 MHz");
	if (mclk > 60e6) die ("MCLK can not be over 60 MHz");

	// unlock flash, if necessary
	r = read_word(MC_FSR, &word);
	if (r) die("error while readying flash status");
	lockmask = (word >> 16) & 0xFFFF;
	if (lockmask) {
		r = unlock_flash_pages(lockmask, mclk);
		if (r) die("Unable to unlock flash pages");
	}

	// change the flash config for writing to flash memory (1.0 us setting)
	// ...rather than lock bits which need a 1.5 us setting
	clks = (int)ceil(mclk * 1.0e-6);
	write_word(MC_FMR, ((clks & 255) << 16) | 0x00000100);

	// todo: configure for the chip we're actually using (maybe read chip id
	// register).  For now, these are just hard coded for the SAM7S64 chip
	numblocks = 512;
	blocksize = 128;
	ramarea = 0x00202000;
	flasharea = 0x00100000;

#if 0
	// load the flasher code into RAM
	r = write_memory(ramarea + blocksize, flasher, FLASHER_SIZE);
	if (r) die("Error loading flasher code");
#endif

	// now program all the blocks
	for (block=0; block<numblocks; block++) {
		// skip any blocks not used by the code
		if (!bytes_within_range(block * blocksize,
		   (block + 1) * blocksize - 1)) {
			//printf("skipping block #%d, not used by code\n", block);
			continue;
		}

		// make sure the flash is ready
		r = wait_flash_ready();
		if (r) die("Timeout waiting for flash ready");

		// write a block into the flash buffer
		for (i=0; i<blocksize; i+=4) {
			r = write_word(flasharea + block * blocksize + i,
				word_from_firmware(block * blocksize + i));
			if (r) die("error writing word\n");
		}

		// and tell the flash controller to burn it into the flash memory!
		r = write_word(MC_FCR, (block << 8) | 0x5A000001);
		if (r) die("error writing flash command\n");

#if 0
		// transfer the data into RAM
		printf("Block #%d\n", block);
		r = write_memory(ramarea, firmware_image + block * blocksize, blocksize);
		if (r) die("error transfering data to ram");

		// write params needed by the flasher code
		r = write_word(ramarea + blocksize, flasharea + block * blocksize);
		if (r) die("error writing flasher param #1 to ram");
		r = write_word(ramarea + blocksize + 4, block);
		if (r) die("error writing flasher param #2 to ram");
		r = write_word(ramarea + blocksize + 8, blocksize);
		if (r) die("error writing flasher param #3 to ram");

		// and jump to the flasher code to do the work
		r = run_code(ramarea + blocksize + 16);
		if (r) die("error running flasher code");
#endif
	}

	return 0;
}


